"""
WavesChainQueryServiceImpl
"""
from waves_gateway.common import map_array, filter_array, Injectable, WAVES_NODE, WAVES_ASSET_ID, WavesNodeException
from waves_gateway.model import TransactionReceiver, Transaction, TransactionSender
from typing import List, Optional

import gevent.pool
from .chain_query_service import ChainQueryService
from doc_inherit import method_doc_inherit  # type: ignore
import requests
import simplejson as json  # type: ignore
import base58  # type: ignore
from .token import WAVES_CHAIN_QUERY_SERVICE


@Injectable(provides=WAVES_CHAIN_QUERY_SERVICE, deps=[WAVES_NODE, WAVES_ASSET_ID])
class WavesChainQueryServiceImpl(ChainQueryService):
    """
    Implements the ChainQueryService for the Waves Blockchain.
    It requires a waves_node that is used for the requests.
    """

    def __init__(self, waves_node: str, asset_id: str) -> None:
        self._waves_node = waves_node
        self._asset_id = asset_id

    def _get_response_from_node(self, path):
        response = requests.get(self._waves_node + path)

        if not response.ok:
            raise WavesNodeException(response.url, response.status_code, response.text)
        else:
            return json.loads(response.text)

    def _convert_node_response_to_transaction(self, res: dict) -> Optional[Transaction]:
        if 'assetId' not in res or res['assetId'] != self._asset_id:
            return None

        if res['type'] != 4:
            return None

        try:
            address = res['recipient'].split(':')[1]
        except IndexError:
            address = res['recipient']
        try:
            sender = res['sender'].split(':')[1]
        except IndexError:
            sender = res['sender']

        return Transaction(
            tx=res['id'],
            receivers=[TransactionReceiver(address=address, amount=res['amount'])],
            senders=[TransactionSender(address=sender)])

    @method_doc_inherit
    def get_transactions_of_block_at_height(self, height: int) -> List[Transaction]:
        transactions = self._get_response_from_node('/blocks/at/' + str(height))['transactions']
        waves_transactions = map_array(self._convert_node_response_to_transaction, transactions)
        return filter_array(lambda x: x is not None, waves_transactions)

    @method_doc_inherit
    def get_height_of_highest_block(self) -> int:
        return self._get_response_from_node('/blocks/height')['height']

    def get_coin_receiver_address_from_transaction(self, transaction: str) -> str:
        """
        Returns the custom currency address that shall be benefited by this transaction
        """
        attachment = self._get_response_from_node('/transactions/info/' + transaction)['attachment']

        return base58.b58decode(attachment).decode("utf-8")

    @method_doc_inherit
    def get_transaction_by_tx(self, tx: str) -> Optional[Transaction]:
        res = self._get_response_from_node('/transactions/info/' + tx)

        if 'status' in res and res['status'] == 'error':
            return None

        return self._convert_node_response_to_transaction(res)



"""
AttemptListWorker
"""

import gevent.greenlet as greenlet
import logging

from waves_gateway.model import PollingDelayConfig
from .transaction_attempt_list_service import TransactionAttemptListService
from .pending_attempt_list_selector_service import PendingAttemptListSelectorService
import gevent


class AttemptListWorker(greenlet.Greenlet):
    """
    This worker is intended to be spawned once or multiple times at the beginning of the application.
    After that, it will reschedule itself in dynamic intervals.
    The interval can never be less than the min_polling_delay and never be more than the max_polling_delay.
    In every iteration the attempt_list_selector is queried for a pending attempt_list.
    This attempt_list will then be processed by this worker.
    """

    def __init__(self,
                 attempt_list_service: TransactionAttemptListService,
                 attempt_list_selector: PendingAttemptListSelectorService,
                 logger: logging.Logger,
                 worker_id: int,
                 min_polling_delay_s: float = PollingDelayConfig.DEFAULT_MIN_ATTEMPT_LIST_WORKER_DELAY_S,
                 max_polling_delay_s: float = PollingDelayConfig.DEFAULT_MAX_ATTEMPT_LIST_WORKER_DELAY_S) -> None:
        greenlet.Greenlet.__init__(self)
        self._attempt_list_service = attempt_list_service
        self._polling_delay = 0
        self._attempt_list_selector = attempt_list_selector
        self._cancelled = False
        self._logger = logger.getChild(AttemptListWorker.__name__ + '[' + str(worker_id) + ']')
        self._delay_s = 0
        self._min_polling_delay_s = min_polling_delay_s
        self._max_polling_delay_s = max_polling_delay_s
        self._first_run_complete = False

    def _log_delay(self):
        self._logger.debug('Sleeping %ss', str(self._delay_s))

    def _run(self):
        try:
            while not self._cancelled or not self._first_run_complete:
                current = None

                try:
                    self._logger.debug('Trying to get new attempt list')
                    current = self._attempt_list_selector.lock_attempt_list()

                    if current is not None:
                        self._attempt_list_service.continue_transaction_attempt_list(current)
                        self._attempt_list_selector.release_attempt_list(current)
                        current = None
                        self._decrease_polling_delay()
                        self._logger.debug('Successfully finished')
                    else:
                        self._logger.debug('Nothing to do')
                        self._increase_polling_delay()
                except BaseException as ex:
                    self._increase_polling_delay()
                    self._logger.error(ex, exc_info=True)
                    self._logger.debug("Execution has finished with errors")
                finally:
                    if current is not None:
                        self._attempt_list_selector.release_attempt_list(current)

                    self._first_run_complete = True
                    self._log_delay()
                    gevent.sleep(self._delay_s)
        finally:
            self._logger.info("Cancelled")

    def _decrease_polling_delay(self):
        self._delay_s = self._delay_s - 1

        if self._delay_s <= self._min_polling_delay_s:
            self._delay_s = self._min_polling_delay_s

    def _increase_polling_delay(self):
        self._delay_s = self._delay_s + 1

        if self._delay_s >= self._max_polling_delay_s:
            self._delay_s = self._max_polling_delay_s

    def cancel(self):
        self._cancelled = True



"""
Collection of common tokens in th Application.
"""

import pywaves
from waves_gateway.model import KeyPair
from .injector import InjectionToken
from pymongo.collection import Collection

LOGGING_HANDLER_LIST = InjectionToken('LOGGING_HANDLER_LIST', list)
MANAGED_LOGGER_LIST = InjectionToken('MANAGED_LOGGER_LIST', list)
POLLING_DELAY_SECONDS = InjectionToken('POLLING_DELAY_SECONDS', float)
CUSTOM_CURRENCY_NAME = InjectionToken('CUSTOM_CURRENCY_NAME', str)
GATEWAY_OWNER_ADDRESS = InjectionToken('GATEWAY_OWNER_ADDRESS', str)
WALLET_STORAGE_COLLECTION_NAME = InjectionToken('WALLET_STORAGE_COLLECTION_NAME', str)
MAP_STORAGE_COLLECTION_NAME = InjectionToken('MAP_STORAGE_COLLECTION_NAME', str)
KEY_VALUE_STORAGE_COLLECTION_NAME = InjectionToken('KEY_VALUE_STORAGE_COLLECTION_NAME', str)
TRANSACTION_ATTEMPT_LIST_STORAGE_COLLECTION_NAME = InjectionToken('TRANSACTION_ATTEMPT_LIST_STORAGE_COLLECTION_NAME',
                                                                  str)
GATEWAY_COIN_ADDRESS_SECRET = InjectionToken('GATEWAY_COIN_ADDRESS_SECRET', KeyPair)
GATEWAY_COIN_ADDRESS = InjectionToken('GATEWAY_COIN_ADDRESS', str)
WAVES_NODE = InjectionToken('WAVES_NODE', str)
COIN_NODE = InjectionToken('COIN_NODE', str)
WAVES_ASSET_ID = InjectionToken('WAVES_ASSET_ID', str)
WAVES_CHAIN = InjectionToken('WAVES_CHAIN', str)
ONLY_ONE_TRANSACTION_RECEIVER = InjectionToken('ONLY_ONE_TRANSACTION_RECEIVER', bool)
COIN_TRANSACTION_WEB_LINK = InjectionToken('COIN_TRANSACTION_WEB_LINK', str)
WAVES_TRANSACTION_WEB_LINK = InjectionToken('WAVES_TRANSACTION_WEB_LINK', str)
COIN_ADDRESS_WEB_LINK = InjectionToken('COIN_ADDRESS_WEB_LINK', str)
WAVES_ADDRESS_WEB_LINK = InjectionToken('WAVES_ADDRESS_WEB_LINK', str)
ATTEMPT_LIST_MAX_COMPLETION_TRIES = InjectionToken('ATTEMPT_LIST_MAX_COMPLETION_TRIES', int)
GATEWAY_WAVES_ADDRESS_SECRET = InjectionToken('GATEWAY_WAVES_ADDRESS_SECRET', KeyPair)
GATEWAY_WAVES_ADDRESS = InjectionToken('GATEWAY_WAVES_ADDRESS', str)
NUM_ATTEMPT_LIST_WORKERS = InjectionToken('NUM_ATTEMPT_LIST_WORKERS', int)
GATEWAY_HOST = InjectionToken('GATEWAY_HOST', str)
GATEWAY_PORT = InjectionToken('GATEWAY_PORT', int)
WALLET_STORAGE_COLLECTION = InjectionToken('WALLET_STORAGE_COLLECTION', Collection)
MAP_STORAGE_COLLECTION = InjectionToken('MAP_STORAGE_COLLECTION', Collection)
KEY_VALUE_STORAGE_COLLECTION = InjectionToken('KEY_VALUE_STORAGE_COLLECTION', Collection)
TRANSACTION_ATTEMPT_LIST_STORAGE_COLLECTION = InjectionToken('TRANSACTION_ATTEMPT_LIST_STORAGE_COLLECTION', Collection)
GATEWAY_PYWAVES_ADDRESS = InjectionToken('GATEWAY_PYWAVES_ADDRESS', pywaves.Address)
COIN_CHAIN_QUERY_SERVICE_CONVERTER_PROXY = InjectionToken('COIN_CHAIN_QUERY_SERVICE_CONVERTER_PROXY')
WAVES_CHAIN_QUERY_SERVICE_CONVERTER_PROXY = InjectionToken('WAVES_CHAIN_QUERY_SERVICE_CONVERTER_PROXY')
FLASK_NAME = InjectionToken('FLASK_NAME', str)
COIN_MAX_HANDLE_TRANSACTION_TRIES = InjectionToken('COIN_MAX_HANDLE_TRANSACTION_TRIES', int)
WAVES_MAX_HANDLE_TRANSACTION_TRIES = InjectionToken('WAVES_MAX_HANDLE_TRANSACTION_TRIES', int)
WAVES_LAST_BLOCK_DISTANCE = InjectionToken('WAVES_LAST_BLOCK_DISTANCE', int)
COIN_LAST_BLOCK_DISTANCE = InjectionToken('COIN_LAST_BLOCK_DISTANCE', int)
WEB_PRIMARY_COLOR = InjectionToken('WEB_PRIMARY_COLOR', str)
WAVES_CHAIN_ID = InjectionToken('WAVES_CHAIN_ID', str)



"""PollingStateStorageProxy"""

from abc import ABC, abstractmethod
from waves_gateway.model import PollingState


class PollingStateStorageProxy(ABC):
    """Offers access to the current polling_state."""

    @abstractmethod
    def get_polling_state(self) -> PollingState:
        pass

    @abstractmethod
    def set_polling_state(self, state: PollingState) -> None:
        pass



"""
AddressValidationService
"""

from abc import ABC, abstractmethod


class AddressValidationService(ABC):
    """
    Allows the validation of an address of a specific cryptocurrency.
    """

    @abstractmethod
    def validate_address(self, address: str) -> bool:
        """
        Should return true or false whether the given string is a valid address.
        """
        pass



"""
Contains common helper classes.
"""

from .exceptions import *
from .utils import convert_to_int, convert_to_decimal, map_array, filter_array
from .proxy_guard import ProxyGuard
from .injector import Injectable, INJECTOR, Factory, InjectionToken, Injector, InjectorError, Token, DependencyConfiguration
from .token import *
from .proxy_factory import ProxyFactory



"""
ChainQueryServiceConverterProxyImpl
"""

from typing import List, Optional

from waves_gateway.common import map_array, Injectable, COIN_CHAIN_QUERY_SERVICE_CONVERTER_PROXY, \
    WAVES_CHAIN_QUERY_SERVICE_CONVERTER_PROXY
from waves_gateway.model import Transaction

from .chain_query_service import ChainQueryService
from .integer_converter_service import IntegerConverterService
from .token import COIN_CHAIN_QUERY_SERVICE, COIN_INTEGER_CONVERTER_SERVICE, ASSET_INTEGER_CONVERTER_SERVICE, WAVES_CHAIN_QUERY_SERVICE


@Injectable(COIN_CHAIN_QUERY_SERVICE_CONVERTER_PROXY, deps=[COIN_CHAIN_QUERY_SERVICE, COIN_INTEGER_CONVERTER_SERVICE])
@Injectable(
    WAVES_CHAIN_QUERY_SERVICE_CONVERTER_PROXY, deps=[WAVES_CHAIN_QUERY_SERVICE, ASSET_INTEGER_CONVERTER_SERVICE])
class ChainQueryServiceConverterProxyImpl(ChainQueryService):
    """
    Applies the given IntegerConverterService on the result of the given ChainQueryService.
    """

    def get_transaction_by_tx(self, tx: str) -> Optional[Transaction]:
        transaction = self._chain_query_service.get_transaction_by_tx(tx)

        if transaction is not None:
            return self._integer_converter_service.convert_transaction_to_int(transaction)
        else:
            return None

    def __init__(self, chain_query_service: ChainQueryService,
                 integer_converter_service: IntegerConverterService) -> None:
        self._integer_converter_service = integer_converter_service
        self._chain_query_service = chain_query_service

    def get_transactions_of_block_at_height(self, height: int) -> List[Transaction]:
        transactions = self._chain_query_service.get_transactions_of_block_at_height(height)
        return map_array(lambda x: self._integer_converter_service.convert_transaction_to_int(x), transactions)

    def get_height_of_highest_block(self) -> int:
        return self._chain_query_service.get_height_of_highest_block()

    def get_coin_receiver_address_from_transaction(self, transaction: str):
        return self._chain_query_service.get_coin_receiver_address_from_transaction(transaction)



"""
Utilities
"""

from decimal import Decimal
from fractions import Fraction
from numbers import Number
from typing import Any, Optional, Iterable
from math import modf


def _parse_value(value: Any) -> Fraction:
    """
    Ensures that the given value has the type Fraction.
    """

    if isinstance(value, Decimal) or isinstance(value, float) or isinstance(value, int):
        return Fraction(value)  # type: ignore
    elif isinstance(value, Fraction):
        return value
    else:
        raise TypeError("The given value is neither a Decimal, float or Fraction")


def _assert_no_fractional(value: Any, msg="is not fractional"):
    if modf(value)[0] != 0.0:
        raise ArithmeticError(msg)


def convert_to_int(value: Any, factor: Any) -> int:
    """
    Converts the given value to an integer by first multiplying it with the given factor and then
    casting the result to an integer. Information may get lost if the factor is too low.
    """
    parsed_v = _parse_value(value)

    parsed_factor = _parse_value(factor)

    _assert_no_fractional(parsed_factor, "factor cannot be a fraction; the factor is " + str(parsed_factor))

    return round(parsed_v * parsed_factor)


def convert_to_decimal(amount: int, factor: Number, precision: Optional[int] = None) -> Decimal:
    """
    Converts the given value to a Decimal by first dividing it by the given factor and then
    casting the result to a Decimal.
    If a precision is provided, round will be called with the result and the given precision.

    """
    parsed_v = _parse_value(amount)
    parsed_factor = _parse_value(factor)

    _assert_no_fractional(parsed_factor, "factor cannot be a fraction; the factor is " + str(parsed_factor))

    res = parsed_v / parsed_factor

    if precision is not None:
        return round(Decimal(float(res)), precision)  # type: ignore
    else:
        return Decimal(float(res))


def map_array(func, arr: list) -> list:
    """
    Calls the given function on every element on the given array and returns the resulting elements.
    """
    res = list()

    for el in arr:
        res.append(func(el))

    return res


def filter_array(func, arr: Iterable) -> list:
    """
    Filters the arr using the given function. The function must return True or False whether
    the element should be part of the result or not.
    """
    res = list()

    for el in arr:
        if func(el):
            res.append(el)

    return res



"""
Exceptions
"""
from typing import Union

from decimal import Decimal


class ParseError(Exception):
    """
    Indicates that a parsing error occurred, for example when parsing JSON or dict.
    """
    pass


class InvalidConfigError(Exception):
    """
    Raised when the provided Gateway configuration is not valid
    and the Gateway can therefore not be constructed.
    """
    pass


class DuplicateMappingError(Exception):
    """
    Raised when the application tries to create a mapping that already exists.
    This should not happen which means that this is an internal error of the framework.
    """
    pass


class DuplicateSecretError(Exception):
    """
    Raised when the application tries to save a secret for the same address multiple times.
    This should not happen which means that this is an internal error of the framework.
    """

    def __init__(self, address: str) -> None:
        super().__init__("Attempted to store secret of address '" + address + "', but it already exists.")


class DuplicateTransactionError(Exception):
    """
    Raised when the application tries to create a transaction that already exists.
    This should not happen which means that this is an internal error of the framework.
    """

    def __init__(self, tx: str) -> None:
        super().__init__("Attempted to store transaction '" + tx + "', but it already exists.")


class MissingSecretError(Exception):
    """
    Raised when the application needs a specific secret, but it is not available.
    """

    def __init__(self, address: str) -> None:
        super().__init__("Did not found secret for address '" + address + "'. Cannot perform transaction.")


class AddressHasTooLessAmountError(Exception):
    """
    Raised when the application attempts to make a transaction, but the sender address has too less amount.
    """

    def __init__(self, expected_amount: Union[int, Decimal], real_amount: Union[int, Decimal], address: str) -> None:
        super().__init__("Expected the address '" + address + "' to contain at least an amount of " +
                         str(expected_amount) + ", but it only has an amount of " + str(real_amount))


class MappingNotFoundForCoinAddress(Exception):
    """
    Raised when a specific mapping was expected to exist, but not found.
    """

    def __init__(self, coin_address: str) -> None:
        super().__init__(
            "The address '" + coin_address + "' has no associated Waves address, but was expected to have one.")


class MultipleGatewayReceiversError(Exception):
    """
    Raised when the Gateway encounters a transaction with multiple receiver addresses.
    """

    def __init__(self, tx) -> None:
        super().__init__(
            "Encountered a transaction (" + tx + ") with multiple receiver addresses. The Gateway cannot handle this.")


class DuplicateAttemptListTriggerError(Exception):
    """
    Raised when trying to create another attempt_list with the same trigger.
    """
    pass


class DuplicateAttemptListIDError(Exception):
    """
    Raised when trying to create another attempt_list with the same identifier.
    """
    pass


class PyWavesError(Exception):
    """
    Raised when PyWaves encountered any error.
    """
    pass


class WavesAddressInvalidError(Exception):
    """
    Raised when an invalid waves address was overhanded.
    """
    pass


class InvalidTransactionIdentifier(Exception):
    """
    Raised when the transaction identifier is not part of the blockchain.
    """
    pass


class InjectorError(Exception):
    """
    Raised when a token could not be resolved for some reason.
    """
    pass


class WavesNodeException(Exception):
    """
    Raised when the waves node was unable to process the request for some reason.
    """

    def __init__(self, url: str, status_code: int, text: str) -> None:
        super().__init__(url + " " + str(status_code) + " " + text)



"""
KeyPair
"""

from typing import Any


class KeyPair(object):
    """
    Pair of a public and private element.
    """
    DICT_PUBLIC_KEY = 'public'
    DICT_SECRET_KEY = 'secret'

    def __init__(self, public: str, secret: str) -> None:
        """

        :param public: The public part is not required to be protected
        :param secret: The secret part must not be leaked
        """
        self._public = public
        self._secret = secret

    @property
    def public(self) -> str:
        return self._public

    @property
    def secret(self) -> str:
        return self._secret

    def __eq__(self, other: Any):
        """
        Compares two KeyPairs for equality.
        """
        if not isinstance(other, KeyPair):
            return False
        else:
            return other.secret == self.secret and other.public == self.public



"""ProxyFactory"""

from typing import Callable


class ProxyFactory(object):
    """
    Create a new instance (using the provided factory) for every attribute request.
    """

    def __init__(self, factory: Callable) -> None:
        self._factory = factory

    def __getattr__(self, item):
        return getattr(self._factory(), item)



"""
Defines services that implement the logic of the Gateway.
"""

from .chain_query_service import ChainQueryService
from .transaction_service import TransactionService
from .transaction_consumer import TransactionConsumer
from .transaction_polling_service import TransactionPollingService
from .integer_converter_service import IntegerConverterService
from .fee_service import FeeService
from .fee_service_converter_proxy_impl import FeeServiceConverterProxyImpl
from .constant_fee_service_impl import ConstantFeeServiceImpl
from .coin_transaction_consumer_impl import CoinTransactionConsumerImpl
from .waves_transaction_consumer_impl import WavesTransactionConsumerImpl
from .waves_chain_query_service_impl import WavesChainQueryServiceImpl
from .asset_transaction_service_impl import AssetTransactionServiceImpl
from .chain_query_service_converter_proxy_impl import ChainQueryServiceConverterProxyImpl
from .transaction_service_converter_proxy_impl import TransactionServiceConverterProxyImpl
from .transaction_service_forwarder_proxy_impl import TransactionServiceForwarderProxyImpl
from .transaction_attempt_list_service import TransactionAttemptListService
from .secret_service import SecretService
from .attempt_list_converter_service import AttemptListConverterService
from .address_validation_service import AddressValidationService
from .waves_address_validation_service import WavesAddressValidationService
from .pending_attempt_list_selector_service import PendingAttemptListSelectorService
from .attempt_list_worker import AttemptListWorker
from .gateway_validation_service import GatewayValidationService
from .gateway_application_service import GatewayApplicationService
from .gateway_logging_configuration_service import GatewayLoggingConfigurationService
from .token import *
from .gateway_config_parser import GatewayConfigParser



"""
WalletStorage
"""

from abc import ABC, abstractmethod
from typing import Optional

from waves_gateway.model import KeyPair
from waves_gateway.common import DuplicateSecretError


class WalletStorage(ABC):
    """
    Storage for private keys.
    """

    @abstractmethod
    def get_secret_by_public_address(self, public_address: str) -> Optional[KeyPair]:
        """
        Returns the secret that is associated to the given public address or None if no such secret exists.
        """
        pass

    def public_address_exists(self, public_address: str) -> bool:
        """
        Returns whether there is an entry with an CoinAddressSecret that has the given public public_address.
        """
        return self.get_secret_by_public_address(public_address) is not None

    @abstractmethod
    def save_address_secret(self, secret: KeyPair) -> None:
        """
        Stores the given address secret.
        """
        pass

    def safely_save_address_secret(self, secret: KeyPair) -> None:
        """
        Before inserting the given secret, the method tests if the public address of the secret is already
        listed in the database. If so, an DuplicateSecretError will be thrown.
        """
        if not self.public_address_exists(secret.public):
            return self.save_address_secret(secret)
        else:
            raise DuplicateSecretError(address=secret.public)



"""
Injector
"""

from typing import List, Any, Dict, Optional, Callable, Union, NewType, Type
from .exceptions import InjectorError
from .utils import map_array


class InjectionToken(object):
    """
    Represents a token that can be used to inject something.
    """

    def __init__(self, name: str, _type: Optional[Any] = None) -> None:
        self._name = name
        self._type = _type

    def __str__(self):
        if self._type is not None:
            return '(' + self._name + ', ' + str(self._type.__name__) + ')'
        else:
            return self._name


Token = Union[Callable, str, InjectionToken]


class DependencyConfiguration(object):
    """
    Configuration for a specific token that is necessary to instantiate the represented class or it may already contain an instance/variable.
    """

    def __init__(self,
                 deps: List[Token] = None,
                 instance: Any = None,
                 factory: Optional[Callable] = None,
                 opt_deps: List[Token] = None,
                 weak: bool = False) -> None:
        if deps is None:
            self._deps = list()  # type: List[Token]
        else:
            self._deps = deps

        if opt_deps is None:
            self._opt_deps = list()  # type: List[Token]
        else:
            self._opt_deps = opt_deps

        self._instance = instance
        self._factory = factory
        self._weak = weak

    @property
    def deps(self) -> List[Token]:
        return self._deps

    @property
    def factory(self) -> Optional[Callable]:
        return self._factory

    @property
    def instance(self) -> Any:
        return self._instance

    @instance.setter
    def instance(self, instance: Any):
        self._instance = instance

    @property
    def opt_deps(self):
        return self._opt_deps

    @property
    def weak(self):
        return self._weak


GLOBAL_DEPENDENCY_MAP = dict()  # type: Dict[Token, DependencyConfiguration]


class Injector(object):
    """
    An injector can be used to retrieve a specific instance or variable by a token.
    It operates on a dependency_map that contains all configuration necessary for the injector to work.
    The root injector uses the global_dependency_map. Any other injector has its own dependency_map and forwards missing tokens to its parent
    injector.
    """

    def __init__(self, dependency_map: Dict[Token, Any]) -> None:
        self._dependency_map = dependency_map

    def provide(self, token: Token, instance: Any, weak=False) -> None:
        if token in self._dependency_map and self._dependency_map[token].weak is False:
            raise InjectorError('Token ' + str(token) + " is already defined and cannot be overwritten.")

        self._dependency_map[token] = DependencyConfiguration(instance=instance, weak=weak)

    def overwrite(self, token: Token, instance: Any) -> None:
        self._dependency_map[token] = DependencyConfiguration(instance=instance)

    def overwrite_if_exists(self, token: Token, instance: Optional[Any]) -> None:
        if instance is not None:
            self.overwrite(token, instance)

    def is_registered(self, token: Token) -> bool:
        """Returns whether the given token is registered in this injector or one of its parents (if any)."""
        return token in self._dependency_map

    def get(self, token: Token) -> Any:
        if not self.is_registered(token):
            raise InjectorError("Cannot provide " + str(token))

        if self._dependency_map[token].instance is not None:
            return self._dependency_map[token].instance

        if self._dependency_map[token].factory is not None:
            self._dependency_map[token].instance = self._dependency_map[token].factory(self)
            return self._dependency_map[token].instance

        try:
            args = map_array(self.get, self._dependency_map[token].deps)
        except InjectorError:
            raise InjectorError('Cannot get all dependencies of ' + str(token))

        def get_opt_dep(token: Token):
            """
            Tries to resolve an dependency and returns None if it has failed.
            """
            try:
                return self.get(token)
            except InjectorError:
                return None

        opt_args = map_array(get_opt_dep, self._dependency_map[token].opt_deps)

        for opt_arg in opt_args:
            args.append(opt_arg)

        if callable(token):
            try:
                self._dependency_map[token].instance = token(*args)
                return self._dependency_map[token].instance
            except InjectorError:
                raise InjectorError('Cannot instantiate ' + str(token))
        else:
            raise InjectorError('Cannot determine how to construct ' + str(token))

    def invoke(self, deps: List[Token], func: Callable, opt_deps: List[Token] = None) -> Any:
        if opt_deps is None:
            opt_deps = list()

        def get_opt_dep(token: Token):
            """Tries to resolve the dependency and returns None of this has failed."""
            try:
                return self.get(token)
            except InjectorError:
                return None

        opt_args = map_array(get_opt_dep, opt_deps)
        args = map_array(self.get, deps)

        for opt_arg in opt_args:
            args.append(opt_arg)

        return func(*args)


INJECTOR = Injector(GLOBAL_DEPENDENCY_MAP)

INJECTOR.provide(Injector, INJECTOR)


class Injectable(object):
    """
    Declares the annotated class to be injectable. It allows to specify a list of dependencies that are injected in the given order.
    The dependencies must be token. In the background this annotation adds an additional token to the global_dependency_map
    making the class usable in any injector instance.
    """

    def __init__(self,
                 provides: Optional[Token] = None,
                 deps: List[Token] = None,
                 dependency_map: Dict[Token, DependencyConfiguration] = GLOBAL_DEPENDENCY_MAP,
                 opt_deps: List[Token] = None,
                 weak: bool = False) -> None:
        if deps is not None:
            self._depends = deps
        else:
            self._depends = list()

        if opt_deps is None:
            self._opt_deps = list()  # type: List[Token]
        else:
            self._opt_deps = opt_deps

        self._token = provides
        self._dependency_map = dependency_map
        self._weak = weak

    def __call__(self, clazz: Type, *args, **kwargs):
        if self._token is not None:
            token = self._token
        else:
            token = clazz

        if token in self._dependency_map and self._dependency_map[token].weak is False:
            raise InjectorError('Token ' + str(token) + " is already defined and cannot be overwritten.")

        self._dependency_map[token] = DependencyConfiguration(
            factory=lambda injector: injector.invoke(self._depends, clazz, self._opt_deps), weak=self._weak)
        return clazz


class Factory(object):
    """
    Declares an annotated function to be a factory for something that is injectable.
    The factory function itself can have dependencies and provides a token.
    """

    def __init__(self,
                 provides: Token,
                 deps: List[Token] = None,
                 opt_deps: List[Token] = None,
                 dependency_map: Dict[Token, DependencyConfiguration] = GLOBAL_DEPENDENCY_MAP,
                 weak: bool = False) -> None:
        if deps is None:
            self._deps = list()  # type: List[Token]
        else:
            self._deps = deps

        if opt_deps is None:
            self._opt_deps = list()  # type: List[Token]
        else:
            self._opt_deps = opt_deps

        self._provides = provides
        self._dependency_map = dependency_map
        self._weak = weak

    def __call__(self, func: Callable, *args, **kwargs):

        if self._provides in self._dependency_map and self._dependency_map[self._provides].weak is False:
            raise InjectorError('Token ' + str(self._provides) + " is already defined and cannot be overwritten.")

        self._dependency_map[self._provides] = DependencyConfiguration(
            factory=lambda injector: injector.invoke(self._deps, func, self._opt_deps), weak=self._weak)

        return func



"""
FeeServiceConverterProxyImpl
"""
from waves_gateway.common import Injectable
from waves_gateway.service.token import COIN_INTEGER_CONVERTER_SERVICE

from .fee_service import FeeService
from .integer_converter_service import IntegerConverterService


@Injectable(deps=[FeeService, COIN_INTEGER_CONVERTER_SERVICE])
class FeeServiceConverterProxyImpl(FeeService):
    """
    Converts the results of the FeeService to int values that can be further processed by the Gateway.
    """

    def __init__(self, fee_service: FeeService, coin_integer_converter_service: IntegerConverterService) -> None:
        self._fee_service = fee_service
        self._coin_converter_service = coin_integer_converter_service

    def get_gateway_fee(self) -> int:
        return self._coin_converter_service.safely_convert_to_int(self._fee_service.get_gateway_fee())

    def get_coin_fee(self) -> int:
        return self._coin_converter_service.safely_convert_to_int(self._fee_service.get_coin_fee())

    def get_waves_fee(self) -> int:
        return self._fee_service.get_waves_fee()



"""
TransactionAttemptListService
"""

import logging

from waves_gateway.common import Injectable
from .transaction_service_forwarder_proxy_impl import TransactionServiceForwarderProxyImpl

from .secret_service import SecretService
from waves_gateway.storage import TransactionAttemptListStorage
from .transaction_service import TransactionService
from waves_gateway.model import TransactionAttemptList, TransactionAttempt


@Injectable(deps=[TransactionServiceForwarderProxyImpl, SecretService, logging.Logger, TransactionAttemptListStorage])
class TransactionAttemptListService(object):
    """
    Defines how to process a TransactionAttemptList.
    """

    def __init__(self, transaction_service: TransactionService, secret_service: SecretService, logger: logging.Logger,
                 transaction_attempt_list_storage: TransactionAttemptListStorage) -> None:
        self._transaction_service = transaction_service
        self._secret_service = secret_service
        self._logger = logger.getChild(self.__class__.__name__)
        self._storage = transaction_attempt_list_storage

    def continue_transaction_attempt_list(self, attempt_list: TransactionAttemptList) -> None:
        self._logger.info("Trying to complete attempt_list '%s'", attempt_list.attempt_list_id)

        if attempt_list.is_complete():
            self._logger.info("attempt_list is already complete")
            return

        for i in range(0, len(attempt_list.transactions)):
            self._log_attempt_alredy_done(attempt_list.attempts[i])

        while not attempt_list.is_complete():
            next_attempt = attempt_list.next_incomplete_attempt()
            transaction = self._transaction_service.send_coin(next_attempt,
                                                              self._secret_service.get_secret_by_address(
                                                                  next_attempt.currency, next_attempt.sender))
            attempt_list.mark_next_attempt_as_complete(next_attempt, transaction.tx)
            self._storage.update_attempt_list(attempt_list)
            self._log_attempt_success(next_attempt)

        self._logger.info("attempt_list '%s' is complete", attempt_list.attempt_list_id)

    def _log_attempt_success(self, attempt: TransactionAttempt):
        for receiver in attempt.receivers:
            self._logger.info("[%s]: Transferred %s from %s to %s", str(attempt.currency), str(receiver.amount),
                              str(attempt.sender), str(receiver.address))

    def _log_attempt_alredy_done(self, attempt: TransactionAttempt):
        for receiver in attempt.receivers:
            self._logger.info("[%s]: Already transferred %s from %s to %s", str(attempt.currency), str(receiver.amount),
                              str(attempt.sender), str(receiver.address))



"""
FlaskRestController
"""

from logging import Logger
from typing import Any, List, Callable, Optional

from waves_gateway.common import map_array, WavesAddressInvalidError, InvalidTransactionIdentifier, Injectable

import flask as flask_module  # type: ignore

from waves_gateway.model.public_configuration import PublicConfiguration
from waves_gateway.serializer import TransactionAttemptListSerializer
from waves_gateway.serializer.public_configuration_serializer import PublicConfigurationSerializer

from .gateway_controller import GatewayController, AttemptListQuery
from .gateway_controller_converter_proxy_impl import GatewayControllerConverterProxyImpl


@Injectable(deps=[
    flask_module.Flask, GatewayControllerConverterProxyImpl, Logger, TransactionAttemptListSerializer,
    PublicConfiguration, PublicConfigurationSerializer
])
class FlaskRestController(object):
    """
    Optional Flask based rest interface for the Gateway.
    Forwards ReST requests to the general GatewayController.
    """

    def _add_url_rule(self, rule: str, endpoint: str, view_func: Callable, methods: List[str] = None):
        self._logger.info('Registered %s %s', str(methods), rule)

        def catch(*args, **kwargs):
            """
            Forwards the request to the concrete handler function.
            """
            try:
                return view_func(*args, **kwargs)
            except BaseException as ex:
                if hasattr(ex, 'code'):
                    raise ex
                else:
                    self._logger.error(ex, exc_info=True)
                    flask_module.abort(500)

        self._flask.add_url_rule(rule, endpoint, catch, methods=methods)

    def __init__(self, flask: Any, gateway_controller: GatewayController, logger: Logger,
                 attempt_list_serializer: TransactionAttemptListSerializer, public_configuration: PublicConfiguration,
                 public_configuration_serializer: PublicConfigurationSerializer) -> None:
        self._flask = flask
        self._gateway_controller = gateway_controller
        self._logger = logger.getChild(self.__class__.__name__)
        self._serializer = attempt_list_serializer
        self._public_configuration = public_configuration
        self._public_configuration_serializer = public_configuration_serializer

        self._add_url_rule('/api/v1', 'heartbeat', lambda: "", methods=['GET'])

        self._add_url_rule(
            '/api/v1/coin-address/<waves_address>',
            'create_address',
            lambda waves_address: self._create_address(waves_address),
            methods=['GET'])

        self._add_url_rule(
            '/api/v1/attempt-list', 'get_attempt_list', lambda: self._get_attempt_list(), methods=['GET'])

        self._add_url_rule(
            '/api/v1/attempt-list/<attempt_list_id>',
            'get_attempt_list_by_id',
            lambda attempt_list_id: self._get_attempt_list_by_id(attempt_list_id),
            methods=['GET'])

        self._add_url_rule('/', 'redirect', lambda: flask_module.redirect('static/'), methods=['GET'])

        self._add_url_rule('/static/', 'index', lambda: flask_module.redirect('static/index.html'), methods=['GET'])

        self._add_url_rule(
            '/api/v1/public-config',
            'get_public_config',
            lambda: flask_module.jsonify(self._public_configuration_serializer.as_dict(self._public_configuration)),
            methods=['GET'])

        self._add_url_rule(
            '/api/v1/check/waves/<tx>', 'check_waves_tx', lambda tx: self._check_waves_transaction(tx), methods=['GET'])

        self._add_url_rule(
            '/api/v1/check/coin/<tx>', 'check_coin_tx', lambda tx: self._check_coin_transaction(tx), methods=['GET'])

    def _create_address(self, waves_address: str) -> str:
        """
        Forwards an create address request to the GatewayController.
        """
        try:
            coin_address = self._gateway_controller.create_address(waves_address)
        except WavesAddressInvalidError:
            return flask_module.abort(400)

        return coin_address

    def _get_attempt_list_by_id(self, attempt_list_id: str) -> Optional[str]:
        attempt_list = self._gateway_controller.get_attempt_list_by_id(attempt_list_id)

        if attempt_list is None:
            return flask_module.abort(404)

        serialized = self._serializer.attempt_list_as_dict(attempt_list)

        return flask_module.jsonify(serialized)

    def _get_attempt_list(self) -> str:
        trigger_tx = flask_module.request.args.get('trigger_tx')  # type: Optional[str]
        trigger_currency = flask_module.request.args.get('trigger_currency')  # type: Optional[str]
        trigger_receiver = flask_module.request.args.get('trigger_receiver')  # type: Optional[str]
        anything = flask_module.request.args.get('anything')  # type: Optional[str]

        trigger_receiver_as_int = None  # type: Optional[int]

        if trigger_receiver is not None:
            trigger_receiver_as_int = int(trigger_receiver)

        if (trigger_tx is None) and (anything is None):
            return flask_module.abort(400)

        query = AttemptListQuery(
            trigger_currency=trigger_currency,
            trigger_tx=trigger_tx,
            trigger_receiver=trigger_receiver_as_int,
            anything=anything)

        attempt_lists = self._gateway_controller.query_attempt_lists(query)

        serialized_lists = map_array(lambda x: self._serializer.attempt_list_as_dict(x), attempt_lists)

        return flask_module.jsonify(serialized_lists)

    def _check_coin_transaction(self, tx: str) -> str:
        try:
            self._gateway_controller.check_coin_transaction(tx)
        except InvalidTransactionIdentifier:
            flask_module.abort(400)
        except NotImplementedError:
            flask_module.abort(501)

        return ''

    def _check_waves_transaction(self, tx: str) -> str:
        try:
            self._gateway_controller.check_waves_transaction(tx)
        except InvalidTransactionIdentifier:
            flask_module.abort(400)

        return ''



"""
AttemptListConverterService
"""

from typing import cast

from waves_gateway.common import Injectable
from waves_gateway.model import TransactionAttempt, TransactionAttemptReceiver, TransactionAttemptList
from waves_gateway.service.token import COIN_INTEGER_CONVERTER_SERVICE, ASSET_INTEGER_CONVERTER_SERVICE

from .integer_converter_service import IntegerConverterService


@Injectable(deps=[COIN_INTEGER_CONVERTER_SERVICE, ASSET_INTEGER_CONVERTER_SERVICE])
class AttemptListConverterService(object):
    """
    Applies the particular converters on an TransactionAttemptList instance.
    """

    def __init__(self, coin_integer_converter_service: IntegerConverterService,
                 asset_integer_converter_service: IntegerConverterService) -> None:
        self._coin_integer_converter_service = coin_integer_converter_service
        self._asset_integer_converter_service = asset_integer_converter_service

    def revert_attempt_conversion(self, attempt: TransactionAttempt):
        if attempt.currency == "coin":
            cv = self._coin_integer_converter_service
        elif attempt.currency == "waves":
            cv = self._asset_integer_converter_service
        else:
            raise Exception("unknown currency")

        receivers = []

        for i in range(0, len(attempt.receivers)):
            amount = cast(int, attempt.receivers[i].amount)  # type: int

            receivers.append(
                TransactionAttemptReceiver(
                    amount=cv.revert_amount_conversion(amount), address=attempt.receivers[i].address))

        fee = attempt.fee

        if attempt.currency == "coin":
            fee = cv.revert_amount_conversion(cast(int, fee))

        return TransactionAttempt(receivers=receivers, fee=fee, sender=attempt.sender, currency=attempt.currency)

    def revert_attempt_list_conversion(self, attempt_list: TransactionAttemptList):
        converted_attempts = []

        for attempt in attempt_list.attempts:
            converted_attempts.append(self.revert_attempt_conversion(attempt))

        return TransactionAttemptList(
            attempts=converted_attempts,
            transactions=attempt_list.transactions,
            trigger=attempt_list.trigger,
            created_at=attempt_list.created_at,
            last_modified=attempt_list.last_modified,
            tries=attempt_list.tries,
            attempt_list_id=attempt_list.attempt_list_id)



"""
TransactionAttemptList
"""

from typing import List, Optional
from uuid import uuid4

import datetime

from .transaction_attempt import TransactionAttempt
from .transaction import TransactionSender


class AttemptListTrigger(object):
    """
    Represents the receiver index of the transaction that has triggered the creation of an
    particular TransactionAttemptList instance.
    """
    DICT_TX = "tx"
    DICT_RECEIVER = "receiver"
    DICT_CURRENCY = "currency"
    DICT_SENDERS_KEY = 'senders'

    def __init__(self, tx: str, receiver: int, currency: str, senders: List[TransactionSender] = None) -> None:
        self._tx = tx
        self._receiver = receiver
        self._currency = currency
        self._senders = senders

    @property
    def tx(self):
        return self._tx

    @property
    def receiver(self):
        return self._receiver

    @property
    def currency(self):
        return self._currency

    @property
    def senders(self):
        return self._senders

    def __eq__(self, other):
        return self.tx == other.tx and self.receiver == other.receiver and self.currency == other.currency


class TransactionAttemptList(object):
    """
    Stores a list of attempts and their corresponding transactions if any.
    An TransactionAttemptList instance defines the progress of processing an incoming transaction.
    The incoming transaction itself is stored as the trigger of the TransactionAttemptList.
    Attempts represent the transactions that are meant to follow the trigger, but they may not already be processed.
    This is expressed by the transactions array in so far that if the transactions array has the same length
    as the attempts array, this means that all attempts have been processed.
    So, each entry in the transactions array is the transaction id of the corresponding TransactionAttempt.
    """
    DICT_ID = "attempt_list_id"
    DICT_TRANSACTIONS = "transactions"
    DICT_ATTEMPTS = "attempts"
    DICT_TRIGGER = "trigger"
    DICT_LAST_MODIFIED = "last_modified"
    DICT_CREATED_AT = "created_at"
    DICT_TRIES = "tries"
    DEFAULT_TRIES = 0

    def __init__(self,
                 trigger: AttemptListTrigger,
                 attempts: List[TransactionAttempt],
                 transactions: Optional[List] = None,
                 created_at: Optional[datetime.datetime] = None,
                 last_modified: Optional[datetime.datetime] = None,
                 tries: int = DEFAULT_TRIES,
                 attempt_list_id: Optional[str] = None) -> None:
        self._trigger = trigger
        self._attempts = attempts
        self._tries = tries
        self._created_at = created_at
        self._last_modified = last_modified

        if transactions is None:
            self._transactions = list()  # type: List[str]
        else:
            self._transactions = transactions

        if attempt_list_id is None:
            self._attempt_list_id = str(uuid4())
        else:
            self._attempt_list_id = attempt_list_id

    @property
    def attempt_list_id(self):
        return self._attempt_list_id

    @property
    def transactions(self) -> List[str]:
        return self._transactions

    @property
    def attempts(self) -> List[TransactionAttempt]:
        return self._attempts

    @property
    def trigger(self):
        return self._trigger

    @property
    def last_modified(self) -> Optional[datetime.datetime]:
        return self._last_modified

    @property
    def created_at(self) -> Optional[datetime.datetime]:
        return self._created_at

    @property
    def tries(self) -> int:
        return self._tries

    def mark_next_attempt_as_complete(self, attempt: TransactionAttempt, tx: str) -> None:
        for i in range(0, len(self._attempts)):
            if self._attempts[i] is attempt:
                if len(self._transactions) != i:
                    raise Exception("Trying to mark an attempt as complete that is not the next attempt")
                self._transactions.append(tx)
                self._last_modified = datetime.datetime.utcnow()

    def increment_tries(self):
        self._tries = self._tries + 1
        self._last_modified = datetime.datetime.utcnow()

    def next_incomplete_attempt(self) -> TransactionAttempt:
        return self._attempts[len(self._transactions)]

    def is_complete(self):
        return len(self._transactions) >= len(self._attempts)

    def __eq__(self, other):
        return self.created_at == other.created_at and self.last_modified == other.last_modified and self.transactions == other.transactions and self.attempts == other.attempts and self.trigger == other.trigger



"""PollingStateSerializer"""

from typing import Dict
from waves_gateway.common import Injectable
from waves_gateway.model import PollingState, PollingTransactionState
from .polling_transaction_state_serializer import PollingTransactionStateSerializer
from .serializer import Serializer


@Injectable(deps=[PollingTransactionStateSerializer])
class PollingStateSerializer(Serializer):
    """Defines how a PollingState can be serialized and deserialized."""

    def __init__(self, polling_transaction_state_serializer: PollingTransactionStateSerializer) -> None:
        self._polling_transaction_state_serializer = polling_transaction_state_serializer

    def _transaction_map_as_dict(self, transaction_map: Dict[str, PollingTransactionState]) -> dict:
        res = dict()

        for transaction in transaction_map:
            res[transaction] = self._polling_transaction_state_serializer.as_dict(transaction_map[transaction])

        return res

    def _transaction_map_from_dict(self, data: dict) -> Dict[str, PollingTransactionState]:
        res = dict()

        for transaction in data:
            res[transaction] = self._polling_transaction_state_serializer.from_dict(data[transaction])

        return res

    def as_dict(self, polling_state: PollingState) -> dict:
        res = dict()

        res[PollingState.DICT_TRANSACTION_MAP_KEY] = self._transaction_map_as_dict(polling_state.transaction_map)

        return res

    def from_dict(self, data: dict) -> PollingState:
        return PollingState(
            transaction_map=self._transaction_map_from_dict(data[PollingState.DICT_TRANSACTION_MAP_KEY]))



"""
MongoTransactionAttemptListStorageImpl
"""

from typing import Optional, List

from waves_gateway.common import Injectable, ATTEMPT_LIST_MAX_COMPLETION_TRIES, \
    TRANSACTION_ATTEMPT_LIST_STORAGE_COLLECTION
from waves_gateway.model import TransactionAttemptList, AttemptListTrigger, AttemptListQuery, TransactionSender
from waves_gateway.storage.transaction_attempt_list_storage import TransactionAttemptListStorage
from waves_gateway.serializer import TransactionAttemptListSerializer
from pymongo.collection import Collection  # type: ignore
from doc_inherit import class_doc_inherit, method_doc_inherit  # type: ignore
import pymongo


@Injectable(
    deps=[
        TRANSACTION_ATTEMPT_LIST_STORAGE_COLLECTION, TransactionAttemptListSerializer, ATTEMPT_LIST_MAX_COMPLETION_TRIES
    ],
    provides=TransactionAttemptListStorage)
class MongoTransactionAttemptListStorageImpl(TransactionAttemptListStorage):
    """
    Implements a storage for TransactionAttemptList instances for a MongoDB database.
    """

    @method_doc_inherit
    def find_by_attempt_list_id(self, attempt_list_id: str) -> Optional[TransactionAttemptList]:
        """inherit"""
        query = dict()
        query[TransactionAttemptList.DICT_ID] = attempt_list_id

        res = self._collection.find_one(query)

        if res is None:
            return None
        else:
            return self._serializer.attempt_list_from_dict(res)

    @method_doc_inherit
    def attempt_list_id_exists(self, attempt_list_id: str) -> bool:
        return self.find_by_attempt_list_id(attempt_list_id) is not None

    @method_doc_inherit
    def find_by_trigger(self, trigger: AttemptListTrigger) -> Optional[TransactionAttemptList]:
        query = self._create_trigger_query(trigger)
        res = self._collection.find_one(query)

        if res is not None:
            return self._serializer.attempt_list_from_dict(res)
        else:
            return None

    def __init__(self,
                 collection: Collection,
                 serializer: TransactionAttemptListSerializer,
                 max_tries: int,
                 limit: int = 10) -> None:
        TransactionAttemptListStorage.__init__(self)
        self._serializer = serializer
        self._collection = collection
        self._limit = limit
        self._max_tries = max_tries

    # noinspection PyMethodMayBeStatic
    def _create_trigger_query(self, trigger: AttemptListTrigger):
        query = dict()

        query[TransactionAttemptList.DICT_TRIGGER + "." + AttemptListTrigger.DICT_RECEIVER] = trigger.receiver
        query[TransactionAttemptList.DICT_TRIGGER + "." + AttemptListTrigger.DICT_CURRENCY] = trigger.currency
        query[TransactionAttemptList.DICT_TRIGGER + "." + AttemptListTrigger.DICT_TX] = trigger.tx

        return query

    @method_doc_inherit
    def save_attempt_list(self, attempt_list: TransactionAttemptList) -> None:
        self._collection.insert_one(self._serializer.attempt_list_as_dict(attempt_list))

    @method_doc_inherit
    def trigger_exists(self, trigger: AttemptListTrigger) -> bool:
        res = self.find_by_trigger(trigger)
        return res is not None

    @method_doc_inherit
    def update_attempt_list(self, attempt_list: TransactionAttemptList) -> None:
        query = self._create_attempt_list_id_query(attempt_list.attempt_list_id)
        self._collection.find_one_and_replace(query, self._serializer.attempt_list_as_dict(attempt_list))

    @method_doc_inherit
    def gateway_transaction_exists(self, tx: str) -> bool:
        query = self._create_transaction_query(tx)
        return self._collection.find_one(query) is not None

    def _create_trigger_tx_query(self, trigger_tx: str) -> dict:
        query = dict()
        query[TransactionAttemptList.DICT_TRIGGER + "." + AttemptListTrigger.DICT_TX] = trigger_tx
        return query

    def _create_trigger_receiver_query(self, trigger_receiver: int) -> dict:
        query = dict()
        query[TransactionAttemptList.DICT_TRIGGER + "." + AttemptListTrigger.DICT_RECEIVER] = trigger_receiver
        return query

    def _create_trigger_currency_query(self, trigger_currency: str) -> dict:
        query = dict()
        query[TransactionAttemptList.DICT_TRIGGER + "." + AttemptListTrigger.DICT_CURRENCY] = trigger_currency
        return query

    def _create_attempt_list_id_query(self, attempt_list_id: str) -> dict:
        query = dict()
        query[TransactionAttemptList.DICT_ID] = attempt_list_id
        return query

    def _create_transaction_query(self, transaction: str) -> dict:
        query = dict()
        query[TransactionAttemptList.DICT_TRANSACTIONS] = transaction
        return query

    def _create_senders_query(self, sender: str) -> dict:
        query = dict()
        subquery = dict()
        subquery[TransactionSender.DICT_ADDRESS_KEY] = sender
        query[TransactionAttemptList.DICT_TRIGGER + '.' + AttemptListTrigger.DICT_SENDERS_KEY] = subquery
        return query

    @method_doc_inherit
    def query_attempt_lists(self, query: AttemptListQuery) -> List[TransactionAttemptList]:
        find_query = dict()  # type: dict

        find_query['$or'] = list()
        or_conditions = find_query['$or']  # type: List[dict]

        if query.trigger_receiver is not None:
            or_conditions.append(self._create_trigger_receiver_query(query.trigger_receiver))

        if query.trigger_currency is not None:
            or_conditions.append(self._create_trigger_currency_query(query.trigger_currency))

        if query.trigger_tx is not None:
            or_conditions.append(self._create_trigger_tx_query(query.trigger_tx))

        if query.anything is not None:
            sub_query = dict()  # type: dict
            sub_query['$or'] = list()
            sub_or_conditions = sub_query['$or']  # type: List[dict]

            sub_or_conditions.append(self._create_trigger_tx_query(query.anything))
            sub_or_conditions.append(self._create_attempt_list_id_query(query.anything))
            sub_or_conditions.append(self._create_transaction_query(query.anything))
            sub_or_conditions.append(self._create_senders_query(query.anything))

            or_conditions.append(sub_query)

        raw_results = self._collection.find(find_query).limit(self._limit)

        results = []

        for raw_result in raw_results:
            results.append(self._serializer.attempt_list_from_dict(raw_result))

        return results

    def find_oldest_pending_attempt_list(self) -> Optional[TransactionAttemptList]:
        query = {
            '$and': [{
                '$nor': [{
                    'attempts': {
                        '$size': 1
                    },
                    'transactions': {
                        '$size': 1
                    }
                }, {
                    'attempts': {
                        '$size': 2
                    },
                    'transactions': {
                        '$size': 2
                    }
                }, {
                    'attempts': {
                        '$size': 3
                    },
                    'transactions': {
                        '$size': 3
                    }
                }]
            }, {
                '$or': [{
                    'tries': {
                        '$lt': self._max_tries
                    }
                }, {
                    'tries': {
                        '$exists': False
                    }
                }]
            }]
        }

        results = list(
            self._collection.find(query).sort(TransactionAttemptList.DICT_LAST_MODIFIED, pymongo.ASCENDING).limit(1))

        if len(results) >= 1:
            return self._serializer.attempt_list_from_dict(results[0])
        else:
            return None



"""
TransactionConsumer
"""

from abc import ABC, abstractmethod

from waves_gateway.model import Transaction


class TransactionConsumer(ABC):
    """
    Consumer of polled transactions
    """

    def filter_transaction(self, transaction: Transaction) -> bool:
        """
        Optional filter that shall be applied on the transactions that are overhanded to the
        handle_transaction method. Allows all transactions by default.
        """
        return True

    @abstractmethod
    def handle_transaction(self, transaction: Transaction) -> None:
        """
        Gets overhanded filtered transactions.
        """
        pass



"""
KeyValueStorage
"""

from abc import ABC, abstractmethod
from typing import Optional
from waves_gateway.model import PollingState


class KeyValueStorage(ABC):
    """
    Storage for key-value pairs.
    """

    @abstractmethod
    def get_last_checked_coin_block_height(self) -> Optional[int]:
        """
        Returns the stored last checked coin block height or None if it was never set.
        """
        pass

    @abstractmethod
    def set_last_checked_coin_block_height(self, block_height: int) -> None:
        """
        Overwrites the last checked coin block height of the custom currency.
        """

    @abstractmethod
    def get_last_checked_waves_block_height(self) -> Optional[int]:
        """
        Returns the stored last checked waves block height or None if it was never set.
        """
        pass

    @abstractmethod
    def set_last_checked_waves_block_height(self, block_height: int) -> None:
        """
        Overwrites the last checked waves block height.
        """

    @abstractmethod
    def set_coin_polling_state(self, polling_state: PollingState) -> None:
        pass

    @abstractmethod
    def get_coin_polling_state(self) -> Optional[PollingState]:
        pass

    @abstractmethod
    def set_waves_polling_state(self, polling_state: PollingState) -> None:
        pass

    @abstractmethod
    def get_waves_polling_state(self) -> Optional[PollingState]:
        pass



"""PollingTransactionStateSerializer"""
from typing import Any, Dict

from waves_gateway.common import Injectable
from waves_gateway.model import PollingTransactionState
from .serializer import Serializer


@Injectable()
class PollingTransactionStateSerializer(Serializer):
    """Defines how a PollingTransactionState can be serialized and deserialized."""

    def as_dict(self, value: PollingTransactionState) -> dict:
        res = dict()  # type: Dict[str, Any]

        res[PollingTransactionState.DICT_OK_KEY] = value.ok
        res[PollingTransactionState.DICT_TRIES_KEY] = value.tries

        return res

    def from_dict(self, data: dict) -> PollingTransactionState:
        return PollingTransactionState(
            ok=data[PollingTransactionState.DICT_OK_KEY], tries=data[PollingTransactionState.DICT_TRIES_KEY])



"""
TransactionServiceForwarderProxyImpl
"""

from typing import Optional

from waves_gateway.common import Injectable
from waves_gateway.model import TransactionAttempt, Transaction
from .token import COIN_TRANSACTION_SERVICE_CONVERTER_PROXY, WAVES_TRANSACTION_SERVICE_CONVERTER_PROXY

from .transaction_service import TransactionService


@Injectable(deps=[COIN_TRANSACTION_SERVICE_CONVERTER_PROXY, WAVES_TRANSACTION_SERVICE_CONVERTER_PROXY])
class TransactionServiceForwarderProxyImpl(TransactionService):
    """
    Forwards the transaction attempt to the correct transaction_service based
    on the selected currency.
    """

    def __init__(self, coin_transaction_service: TransactionService,
                 asset_transaction_service: TransactionService) -> None:
        self._coin_transaction_service = coin_transaction_service
        self._asset_transaction_service = asset_transaction_service

    def send_coin(self, attempt: TransactionAttempt, secret: Optional[str]) -> Transaction:
        if attempt.currency == "waves":
            return self._asset_transaction_service.send_coin(attempt, secret)
        elif attempt.currency == "coin":
            return self._coin_transaction_service.send_coin(attempt, secret)
        else:
            raise Exception("Cannot handle currency " + attempt.currency)



"""
MappingEntry
"""


class MappingEntry(object):
    """
    Defines what will be saved in the MapStorage.
    A MappingEntry represents a connection between a custom currency address and a Waves address.
    """

    DICT_COIN_KEY = 'coin'
    DICT_WAVES_KEY = 'waves'

    def __init__(self, waves_address: str, coin_address: str) -> None:
        self._waves_address = waves_address
        self._coin_address = coin_address

    @property
    def waves_address(self) -> str:
        return self._waves_address

    @property
    def coin_address(self) -> str:
        return self._coin_address

    def __str__(self):
        return "Coin(" + str(self._coin_address) + ")" + " -> " + "Waves(" + str(self._waves_address) + ")"

    def __eq__(self, other):
        return self.waves_address == other.waves_address and self.coin_address == other.coin_address



"""
TransactionService
"""

from abc import ABC, abstractmethod
from typing import Optional

from waves_gateway.model import Transaction, TransactionAttempt


class TransactionService(ABC):
    """
    Defines the functionality of a service that can be used to perform a specific TransactionAttempt.
    """

    @abstractmethod
    def send_coin(self, attempt: TransactionAttempt, secret: Optional[str]) -> Transaction:
        """
        Sends a specific amount of coins and returns the resulting transaction.
        The sender address/account looses (amount + fee) coins.
        """
        pass



"""
MongoWalletStorageImpl
"""

from typing import Optional

from pymongo.collection import Collection  # type: ignore
from doc_inherit import method_doc_inherit  # type: ignore

from waves_gateway.common import Injectable, WALLET_STORAGE_COLLECTION
from waves_gateway.model import KeyPair
from waves_gateway.serializer import KeyPairSerializer
from .wallet_storage import WalletStorage


@Injectable(deps=[WALLET_STORAGE_COLLECTION, KeyPairSerializer], provides=WalletStorage)
class MongoWalletStorageImpl(WalletStorage):
    """
    Wallet storage implementation using a MongoDB.
    """

    def __init__(self, collection: Collection, serializer: KeyPairSerializer) -> None:
        self._collection = collection
        self._serializer = serializer

    @method_doc_inherit
    def save_address_secret(self, secret: KeyPair) -> None:
        self._collection.insert_one(self._serializer.as_dict(secret))

    @method_doc_inherit
    def get_secret_by_public_address(self, public_address: str) -> Optional[KeyPair]:
        query = dict()
        query[KeyPair.DICT_PUBLIC_KEY] = public_address
        query_result = self._collection.find_one(query)

        if query_result is None:
            return None
        else:
            return self._serializer.from_dict(query_result)



"""
MapStorage
"""

from abc import ABC, abstractmethod
from typing import Optional

from waves_gateway.model import MappingEntry
from waves_gateway.common import DuplicateMappingError
import gevent.lock


class MapStorage(ABC):
    """
    Storage that is capable of saving querying instances of MappingEntry.
    """

    def __init__(self):
        self._lock = gevent.lock.Semaphore()

    def coin_address_exists(self, coin_address: str) -> bool:
        """
        Tests if the given coin_address does belong to an mapping.
        This method may be overwritten if a more performant variant exists.
        """
        return self.get_waves_address_by_coin_address(coin_address) is not None

    @abstractmethod
    def get_waves_address_by_coin_address(self, coin_address: str) -> Optional[str]:
        """
        Returns the Waves address that is associated to the given custom currency address.
        May return none, if no such mapping exists.
        """
        pass

    @abstractmethod
    def get_coin_address_by_waves_address(self, coin_address: str) -> Optional[str]:
        """
        Returns the custom currency address that is associated to the given Waves address.
        May return none, if no such mapping exists.
        """
        pass

    def waves_address_exists(self, waves_address: str) -> bool:
        """
        Tests if the given Waves address exists.
        This method may be overwritten if a more performant variant exists.
        """
        return self.get_coin_address_by_waves_address(waves_address) is not None

    @abstractmethod
    def save_mapping(self, mapping: MappingEntry):
        """
        Stores the given mapping.
        """
        pass

    def safely_save_mapping(self, mapping: MappingEntry):
        """
        Checks if the given mapping exists, before it is saved.
        If the mapping does already exist, an DuplicateMappingError is thrown.
        """
        self._lock.acquire()

        try:
            if self.waves_address_exists(mapping.waves_address) or self.coin_address_exists(mapping.coin_address):
                raise DuplicateMappingError()

            res = self.save_mapping(mapping)

            self._lock.release()
        finally:
            self._lock.release()

        return res



"""
ConstantFeeServiceImpl
"""

from typing import Optional, Union

from decimal import Decimal

from .fee_service import FeeService


class ConstantFeeServiceImpl(FeeService):
    """
    Implements a FeeService that uses constant values.
    """

    def __init__(self, gateway_fee: Union[int, Decimal], coin_fee: Union[int, Decimal],
                 waves_fee: Optional[int] = None) -> None:
        self._gateway_fee = gateway_fee
        self._coin_fee = coin_fee
        self._waves_fee = waves_fee

    def get_gateway_fee(self) -> Union[int, Decimal]:
        return self._gateway_fee

    def get_coin_fee(self) -> Union[int, Decimal]:
        return self._coin_fee

    def get_waves_fee(self) -> int:
        if self._waves_fee is not None:
            return self._waves_fee
        else:
            return super().get_waves_fee()



"""
WavesTransactionConsumerImpl
"""

from logging import Logger
from typing import List, cast

import datetime
import pywaves  # type: ignore

from waves_gateway.common import MultipleGatewayReceiversError, Injectable, GATEWAY_OWNER_ADDRESS, \
    GATEWAY_COIN_ADDRESS_SECRET, ONLY_ONE_TRANSACTION_RECEIVER, GATEWAY_WAVES_ADDRESS_SECRET, GATEWAY_PYWAVES_ADDRESS, \
    WAVES_CHAIN_QUERY_SERVICE_CONVERTER_PROXY
from waves_gateway.model import TransactionReceiver, \
    TransactionAttempt, TransactionAttemptReceiver, TransactionAttemptList, AttemptListTrigger, \
    Transaction, TransactionSender, KeyPair
from .fee_service_converter_proxy_impl import FeeServiceConverterProxyImpl

from .token import COIN_ADDRESS_VALIDATION_SERVICE
from waves_gateway.storage import TransactionAttemptListStorage

from .chain_query_service import ChainQueryService
from .address_validation_service import AddressValidationService
from .fee_service import FeeService
from .transaction_attempt_list_service import TransactionAttemptListService
from .transaction_consumer import TransactionConsumer


@Injectable(deps=[
    GATEWAY_WAVES_ADDRESS_SECRET, WAVES_CHAIN_QUERY_SERVICE_CONVERTER_PROXY, Logger, GATEWAY_OWNER_ADDRESS,
    GATEWAY_COIN_ADDRESS_SECRET, FeeServiceConverterProxyImpl, ONLY_ONE_TRANSACTION_RECEIVER,
    TransactionAttemptListService, GATEWAY_PYWAVES_ADDRESS, TransactionAttemptListStorage,
    COIN_ADDRESS_VALIDATION_SERVICE
])
class WavesTransactionConsumerImpl(TransactionConsumer):
    """
    Takes the received transactions from Waves and issues the resulting transactions
    in the custom cryptocurrency.
    """

    def __init__(self, gateway_waves_address_secret: KeyPair, waves_chain_query_service: ChainQueryService,
                 logger: Logger, gateway_owner_address: str, gateway_coin_address_secret: KeyPair,
                 fee_service: FeeService, only_one_transaction_receiver: bool,
                 attempt_service: TransactionAttemptListService, gateway_pywaves_address: pywaves.Address,
                 attempt_list_storage: TransactionAttemptListStorage,
                 coin_address_validation_service: AddressValidationService) -> None:
        self._gateway_waves_address_secret = gateway_waves_address_secret
        self._waves_chain_query_service = waves_chain_query_service
        self._logger = logger.getChild(self.__class__.__name__)
        self._gateway_owner_address = gateway_owner_address
        self._fee_service = fee_service
        self._gateway_coin_holder_secret = gateway_coin_address_secret
        self._only_one_transaction_receiver = only_one_transaction_receiver
        self._attempt_service = attempt_service
        self._gateway_pywaves_address = gateway_pywaves_address
        self._attempt_list_storage = attempt_list_storage
        self._coin_address_validation_service = coin_address_validation_service

    def filter_transaction(self, transaction: Transaction) -> bool:
        """
        Ensures that only transactions are handled that are
        addressed to the gateway main address.
        """
        if self._attempt_list_storage.gateway_transaction_exists(transaction.tx):
            return False
        else:
            return len(self._filter_receivers(transaction)) > 0

    def _filter_receivers(self, transaction: Transaction) -> List[int]:
        res = list()

        for i in range(0, len(transaction.receivers)):
            if self._gateway_pywaves_address.address == transaction.receivers[i].address:
                attempt_list = self._attempt_list_storage.find_by_trigger(
                    AttemptListTrigger(tx=transaction.tx, currency="waves", receiver=i))

                if attempt_list is None or not attempt_list.is_complete():
                    res.append(i)

        return res

    def _handle_receiver(self,
                         tx: str,
                         receiver: TransactionReceiver,
                         index: int,
                         senders: List[TransactionSender] = None) -> None:

        # ---------- Pre-Calculation ----------

        # fee to be used for the custom currency
        coin_fee = cast(int, self._fee_service.get_coin_fee())

        # fee to be transferred to the Gateway owner
        gateway_fee = cast(int, self._fee_service.get_gateway_fee())

        received_amount = cast(int, receiver.amount)

        # the amount of coins to be transferred after applying the fees
        if self._only_one_transaction_receiver:
            amount_after_fees = received_amount - 2 * coin_fee - gateway_fee
        else:
            amount_after_fees = received_amount - coin_fee - gateway_fee

        # receiver coin address
        coin_receiver = self._waves_chain_query_service.get_coin_receiver_address_from_transaction(tx)

        # ---------- Pre-Check ----------

        if not self._coin_address_validation_service.validate_address(coin_receiver):
            self._logger.warning("Received transaction %s with an invalid coin receiver address %s. "
                                 "Will be skipped.", tx, coin_receiver)
            return

        if self._gateway_coin_holder_secret.public == coin_receiver:
            self._logger.warning("Received transaction %s with the address of the Gateway coin holder."
                                 "Will be skipped as an transaction to itself has no effect.", tx)
            return

        # transaction must have at least a specific amount to be processed
        if amount_after_fees < 0:
            self._logger.warning("Received transaction %s with an amount of %s, but it is less than the at least "
                                 "required amount. Will be skipped.", tx, receiver.amount)
            return

        attempt_list = self._attempt_list_storage.find_by_trigger(
            AttemptListTrigger(tx=tx, currency="waves", receiver=index))

        if attempt_list is None:
            trigger = AttemptListTrigger(tx=tx, receiver=index, currency="waves", senders=senders)

            attempts = list()

            if self._only_one_transaction_receiver:
                attempts.append(
                    TransactionAttempt(
                        sender=self._gateway_coin_holder_secret.public,
                        fee=coin_fee,
                        currency="coin",
                        receivers=[TransactionAttemptReceiver(address=self._gateway_owner_address,
                                                              amount=gateway_fee)]))

                attempts.append(
                    TransactionAttempt(
                        sender=self._gateway_coin_holder_secret.public,
                        fee=coin_fee,
                        currency="coin",
                        receivers=[TransactionAttemptReceiver(address=coin_receiver, amount=amount_after_fees)]))
            else:
                attempts.append(
                    TransactionAttempt(
                        sender=self._gateway_coin_holder_secret.public,
                        fee=coin_fee,
                        currency="coin",
                        receivers=[
                            TransactionAttemptReceiver(address=coin_receiver, amount=amount_after_fees),
                            TransactionAttemptReceiver(address=self._gateway_owner_address, amount=gateway_fee)
                        ]))

            attempt_list = TransactionAttemptList(
                trigger, attempts, last_modified=datetime.datetime.utcnow(), created_at=datetime.datetime.utcnow())

            self._attempt_list_storage.safely_save_attempt_list(attempt_list)
            self._logger.info('Created new attempt list %s', str(attempt_list.attempt_list_id))

    def handle_transaction(self, transaction: Transaction) -> None:
        """
        Fetches the custom currency address which should be the receiver.
        Issues the transaction to the custom currency address.
        """

        receivers = self._filter_receivers(transaction)

        if len(receivers) > 1:
            raise MultipleGatewayReceiversError(transaction.tx)

        for i in receivers:
            self._handle_receiver(transaction.tx, transaction.receivers[i], i, transaction.senders)



"""
WavesBlockHeightStorageProxyImpl
"""

from typing import Optional
from doc_inherit import method_doc_inherit  # type: ignore

from waves_gateway.common import Injectable
from waves_gateway.storage.key_value_storage import KeyValueStorage
from waves_gateway.storage.block_height_storage_proxy import BlockHeightStorageProxy


@Injectable(deps=[KeyValueStorage])
class WavesBlockHeightStorageProxyImpl(BlockHeightStorageProxy):
    """
    Forwards requests for the block height to the KeyValueStorage.
    """

    def __init__(self, key_value_storage: KeyValueStorage) -> None:
        self._key_value_storage = key_value_storage

    @method_doc_inherit
    def set_last_checked_block_height(self, block_height: int) -> None:
        return self._key_value_storage.set_last_checked_waves_block_height(block_height)

    @method_doc_inherit
    def get_last_checked_block_height(self) -> Optional[int]:
        return self._key_value_storage.get_last_checked_waves_block_height()



"""
CoinAddressFactory
"""

from abc import ABC, abstractmethod
from typing import Union

from waves_gateway.model import KeyPair


class CoinAddressFactory(ABC):
    """
    Offers the functionality to create new addresses in the custom currency.
    """

    @abstractmethod
    def create_address(self) -> Union[KeyPair, str]:
        """
        Create a new address.
        :return: The return type depends on the existence of an account. If there is an account, it may
        be sufficient to just return the public address. If not, the secret must be returned.
        """
        pass



"""
FeeService
"""

from abc import ABC, abstractmethod
from typing import Union

import pywaves  # type: ignore
from decimal import Decimal


class FeeService(ABC):
    """
    Defines the fee values, the Gateway requires. The waves_fee may also be overwritten of required.
    """

    @abstractmethod
    def get_coin_fee(self) -> Union[int, Decimal]:
        pass

    @abstractmethod
    def get_gateway_fee(self) -> Union[int, Decimal]:
        pass

    def get_waves_fee(self) -> int:
        return pywaves.DEFAULT_TX_FEE



"""
ProxyGuard
"""

from typing import Optional, Any
from gevent.lock import Semaphore


class ProxyGuard(object):
    """
    Helper class that can be used to protect an proxy object that represents a remote resource.
    It is possible to specify the maximum number of parallel calls of any method.
    This is helpful in the case that the resource limits the number of parallel connections.
    """

    def __init__(self,
                 proxy: Any,
                 max_parallel_access=1,
                 attr_name: Optional[str] = None,
                 semaphore: Optional[Semaphore] = None) -> None:
        self._attr_name = attr_name
        self._proxy = proxy
        self._max_parallel_access = max_parallel_access

        if semaphore is not None:
            self._semaphore = semaphore
        else:
            self._semaphore = Semaphore(value=max_parallel_access)

    def __getattr__(self, name):
        return ProxyGuard(
            self._proxy, max_parallel_access=self._max_parallel_access, attr_name=name, semaphore=self._semaphore)

    def __call__(self, *args, **kwargs) -> Any:
        self._semaphore.acquire(blocking=True)

        ex = None
        result = None

        try:
            result = getattr(self._proxy, self._attr_name)(*args, **kwargs)
        except BaseException as base_exception:
            ex = base_exception
        finally:
            self._semaphore.release()

        if ex is not None:
            raise ex

        return result



"""
AssetTransactionServiceImpl
"""

from typing import Optional

from waves_gateway.common import PyWavesError, Injectable, WAVES_ASSET_ID, GATEWAY_PYWAVES_ADDRESS
from waves_gateway.model import TransactionReceiver, TransactionAttempt, Transaction
from .transaction_service import TransactionService
import pywaves as pw  # type: ignore


@Injectable(deps=[WAVES_ASSET_ID, GATEWAY_PYWAVES_ADDRESS])
class AssetTransactionServiceImpl(TransactionService):
    """
    Implements a TransactionService that is capable of processing an TransactionAttempt instance
    with a 'waves' currency.
    """

    def __init__(self, waves_asset_id: str, gateway_pywaves_address: pw.Address) -> None:
        self._w_ltc = pw.Asset(waves_asset_id)
        self._gateway_pywaves_address = gateway_pywaves_address

    def send_coin(self, attempt: TransactionAttempt, secret: Optional[str]) -> Transaction:
        if attempt.sender != self._gateway_pywaves_address.address:
            raise PyWavesError('Missing secret for sender ' + attempt.sender)

        transaction = self._gateway_pywaves_address.sendAsset(
            recipient=pw.Address(attempt.receivers[0].address),
            asset=self._w_ltc,
            amount=attempt.receivers[0].amount,
            txFee=attempt.fee)

        if transaction is None:
            raise PyWavesError("Encountered an unknown exception while trying to perform waves transaction.")

        return Transaction(
            tx=transaction['id'],
            receivers=[TransactionReceiver(transaction['recipient'], attempt.receivers[0].amount)])
